import importlib
import logging
import traceback
from typing import Annotated

import typer
from rich.padding import Padding

from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.utils import exit_with_warning_msg, out_console

logger = logging.getLogger(__name__)


PLUGIN_GROUP = "jobflow_remote.plugins"
PLUGIN_LOAD_FUNCTION = "setup_jf_plugin"

app_plugin = JFRTyper(
    name="plugin",
    help="Commands for managing plugins",
    no_args_is_help=True,
)
app.add_typer(app_plugin)


def load_plugins():
    """
    Load and register plugins from entry points.

    Main function used to load the plugin that will be added to the main CLI
    """

    try:
        for entry_point in importlib.metadata.entry_points().select(group=PLUGIN_GROUP):
            try:
                plugin_module = entry_point.load()

                # Call the plugin's setup function
                if hasattr(plugin_module, PLUGIN_LOAD_FUNCTION):
                    plugin_module.setup_jf_plugin()

            except Exception:
                logging.debug(
                    f"Failed to load plugin {entry_point.name}", exc_info=True
                )

    except Exception:
        logging.debug("Error discovering plugins", exc_info=True)


@app_plugin.command(name="list")
def list_plugins(
    error: Annotated[
        bool,
        typer.Option(
            "--error",
            "-e",
            help="Print the error messages for the plugin that cannot be loaded",
        ),
    ] = False,
):
    """List all plugins with an entry point"""
    from jobflow_remote import SETTINGS

    plugins = {}
    error_msgs = {}
    # if there is an error discovering plugins let it pass as it should
    # be logged by the CLI
    for entry_point in importlib.metadata.entry_points().select(group=PLUGIN_GROUP):
        try:
            plugin_module = entry_point.load()

            # Call the plugin's setup function
            if hasattr(plugin_module, PLUGIN_LOAD_FUNCTION):
                plugin_module.setup_jf_plugin()
                plugins[entry_point.name] = entry_point.value
            else:
                error_msgs[entry_point.name] = f"No {PLUGIN_LOAD_FUNCTION} function"

        except Exception:
            error_msgs[entry_point.name] = traceback.format_exc()

    if plugins:
        out_console.print("Available plugins:")
        for name, module in plugins.items():
            out_console.print(f" - {name}: {module}")
    else:
        out_console.print("No plugins found.")

    if error_msgs:
        if error:
            out_console.print("Errors discovering plugins:")
            for name, error_msg in error_msgs.items():
                out_console.print(f" - {name}:")
                out_console.print(Padding(error_msg, pad=(0, 0, 0, 4)))
        else:
            msg = "Errors discovering plugins."
            if SETTINGS.cli_suggestions:
                msg += " Run the command with the --error option to get the details"
            exit_with_warning_msg(msg)



from __future__ import annotations

import io
import re
from typing import IO, TYPE_CHECKING, Any

from rich.console import Console
from rich.text import Text
from typer.testing import CliRunner, Result

from jobflow_remote.cli.jf import app

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


ansi_esc_pattern = re.compile(r"\x1B\[\d+(;\d+){0,2}m")


def run_check_cli(
    cli_args: str | Sequence[str] | None = None,
    cli_input: bytes | str | IO[Any] | None = None,
    cli_env: Mapping[str, str] | None = None,
    catch_exceptions: bool = False,
    required_out: str | Sequence[str] | None = None,
    excluded_out: str | Sequence[str] | None = None,
    required_out_colored: str | Sequence[str] | None = None,
    error: bool = False,
    terminal_width: int = 1000,
) -> Result:
    if isinstance(required_out, str):
        required_out = [required_out]
    if isinstance(excluded_out, str):
        excluded_out = [excluded_out]
    if isinstance(required_out_colored, str):
        required_out_colored = [required_out_colored]

    cli_runner = CliRunner()

    result = cli_runner.invoke(
        app,
        args=cli_args,
        input=cli_input,
        env=cli_env,
        catch_exceptions=catch_exceptions,
        terminal_width=terminal_width,
        color=True,
    )

    # Since typer 0.16.0 and click 8.2 the stderr is necessarily separated from the output.
    # To keep backward compatibility with previous click versions handle both defaults cases.
    try:
        result_out = result.stdout + "\n" + result.stderr
    except ValueError:
        result_out = result.stdout

    # note that stderr is not captured separately
    assert (
        error == (result.exit_code != 0)
    ), f"cli should have {'' if error else 'not '}failed. exit code: {result.exit_code}. stdout: {result_out}"

    # The print of the output in the console during the tests may result in newlines added
    # that prevent the output to be matched. replace all spaces with a single space.
    single_space_output = re.sub(r"\s+", " ", result_out)
    # Remove ansi escape codes
    single_space_output_no_ansi_esc = ansi_esc_pattern.sub("", single_space_output)
    # Remove again multiple spaces. This is needed otherwise we can sometimes (??) have 2 spaces
    # when ansi escape codes have been removed)
    single_space_output_no_ansi_esc = re.sub(
        r"\s+", " ", single_space_output_no_ansi_esc
    )

    if required_out:
        for ro in required_out:
            assert re.sub(r"\s+", " ", ro) in repr(
                single_space_output_no_ansi_esc
            ), f"{ro} missing from stdout: {result_out}"

    if excluded_out:
        for eo in excluded_out:
            assert (
                re.sub(r"\s+", " ", eo) not in single_space_output_no_ansi_esc
            ), f"{eo} present in stdout: {result_out}"

    if required_out_colored:
        for roc in required_out_colored:
            roc_text = Text.from_markup(roc)
            buf = io.StringIO()
            console = Console(file=buf, force_terminal=True, color_system="truecolor")
            console.print(roc_text, end="")
            roc_string = buf.getvalue()
            assert (
                re.sub(r"\s+", " ", roc_string) in single_space_output
            ), f'"{roc_string}" ({roc}) missing from stdout: {result_out}'

    return result



from __future__ import annotations

import contextlib
import os
import uuid
from collections.abc import Mapping, MutableMapping
from copy import deepcopy
from datetime import datetime, timezone
from typing import Any
from uuid import UUID

import maggma.stores  # required to enable subclass searching
from dateutil.relativedelta import relativedelta
from dateutil.tz import gettz
from maggma.core.store import Store
from monty.json import MontyDecoder


def deep_merge_dict(
    d1: MutableMapping,
    d2: Mapping,
    path: list[str] | None = None,
    raise_on_conflicts: bool = True,
    inplace: bool = True,
) -> MutableMapping:
    """
    Merge a dictionary d2 into a dictionary d1 recursively.

    Parameters
    ----------
    d1
    d2
    path
    raise_on_conflicts
    inplace

    Returns
    -------

    """
    if not inplace:
        d1 = deepcopy(d1)
    if path is None:
        path = []
    for key in d2:
        if key in d1:
            if isinstance(d1[key], Mapping) and isinstance(d2[key], Mapping):
                deep_merge_dict(d1[key], d2[key], [*path, str(key)])
            elif d1[key] == d2[key]:
                pass  # same leaf value
            elif raise_on_conflicts:
                raise ValueError(f"Conflict at {'.'.join([*path, str(key)])}")
            else:
                d1[key] = d2[key]
        else:
            d1[key] = d2[key]
    return d1


def remove_none(obj):
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(remove_none(x) for x in obj if x is not None)
    if isinstance(obj, dict):
        return type(obj)(
            (remove_none(k), remove_none(v))
            for k, v in obj.items()
            if k is not None and v is not None
        )
    return obj


def check_dict_keywords(obj: Any, keywords: list[str]) -> bool:
    if isinstance(obj, (list, tuple, set)):
        return any(check_dict_keywords(x, keywords) for x in obj)
    if isinstance(obj, dict):
        for k, v in obj.items():
            if isinstance(k, str) and any(k.startswith(kw) for kw in keywords):
                return True
            if check_dict_keywords(v, keywords):
                return True
    return False


def uuid_to_path(
    uuid: str, index: int | None = 1, num_subdirs: int = 3, subdir_len: int = 2
):
    u = UUID(uuid)
    u_hex = u.hex

    # Split the digest into groups of "subdir_len" characters
    subdirs = [
        u_hex[i : i + subdir_len]
        for i in range(0, num_subdirs * subdir_len, subdir_len)
    ]

    # add the index to the final dir name
    dir_name = f"{uuid}"
    if index is not None:
        dir_name += f"_{index}"

    # Combine root directory and subdirectories to form the final path
    return os.path.join(*subdirs, dir_name)


def store_from_dict(store_dict: dict) -> Store:
    if "@class" in store_dict and "@module" in store_dict:
        store = MontyDecoder().process_decoded(store_dict)
        if not isinstance(store, Store):
            raise ValueError(
                f"The converted object {store} is not an instance of a maggma Store"
            )
        return store

    def all_subclasses(cl):
        return set(cl.__subclasses__()).union(
            [s for c in cl.__subclasses__() for s in all_subclasses(c)]
        )

    all_stores = {s.__name__: s for s in all_subclasses(maggma.stores.Store)}
    return convert_store(store_dict, all_stores)


def convert_store(spec_dict: dict, valid_stores) -> Store:
    """
    Build a store based on the dict spec configuration from JobFlow
    TODO expose the methods from jobflow and don't duplicate the code.
    """
    _spec_dict = dict(spec_dict)
    store_type = _spec_dict.pop("type")
    for k, v in _spec_dict.items():
        if isinstance(v, dict) and "type" in v:
            _spec_dict[k] = convert_store(v, valid_stores)
    return valid_stores[store_type](**_spec_dict)


def convert_utc_time(datetime_value: datetime) -> datetime:
    """
    Convert a time in UTC (used in the DB) to the time zone of the
    system where the code is being executed.

    Parameters
    ----------
    datetime_value
        a datetime object in UTC

    Returns
    -------
        The datetime in the zone of the current system
    """
    return datetime_value.replace(tzinfo=timezone.utc).astimezone(tz=None)


def get_past_time_rounded(
    interval: str, num_intervals: int, reference: datetime | None = None
) -> datetime:
    """
    Return a datetime object that is the specified number of intervals in the
    past relative to the given reference datetime. The returned datetime is
    rounded to the nearest interval start time.

    Parameters
    ----------
    interval
        One of 'hours', 'days', 'weeks', 'months', 'years'
    num_intervals
        The number of intervals to go back in time
    reference
        The datetime to use as the reference for the calculation. If not
        specified, the current time is used.

    Returns
    -------
    datetime
        The datetime object that is the specified number of intervals in the
        past relative to the given reference datetime.
    """
    if not reference:
        reference = datetime.utcnow()
    past = reference - relativedelta(**{interval: num_intervals - 1})  # type: ignore[arg-type]

    # Define starting point modifications based on quantity type
    start_modifiers = {
        "hours": lambda dt: dt.replace(minute=0, second=0, microsecond=0),
        "days": lambda dt: dt.replace(hour=0, minute=0, second=0, microsecond=0),
        "weeks": lambda dt: dt.replace(hour=0, minute=0, second=0, microsecond=0)
        - relativedelta(days=dt.weekday()),
        "months": lambda dt: dt.replace(
            day=1, hour=0, minute=0, second=0, microsecond=0
        ),
        "years": lambda dt: dt.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0
        ),
    }

    return start_modifiers[interval](past)


def get_utc_offset(timezone: str):
    # Get current time in the specified timezone
    """
    Return the UTC offset of the given timezone as a string.

    Parameters
    ----------
    timezone
        The timezone for which to get the UTC offset.

    Returns
    -------
    str
        The UTC offset as a string in the format +/-HH:MM.
    """
    tz_info_val = gettz(timezone)
    if not tz_info_val:
        raise ValueError(f"Could not determine the timezone for {timezone}")
    now = datetime.now(tz_info_val)

    # Get the UTC offset
    utc_offset = now.utcoffset()

    # Extract hours and minutes
    hours, remainder = divmod(utc_offset.total_seconds(), 3600)
    minutes = remainder // 60

    return f"{int(hours):+03d}:{int(minutes):02d}"


# TODO imported this from jobflow remote for backward compatibility.
# remove this in favor of suid in the next release
def suuid() -> str:
    """
    Generate a string UUID (universally unique identifier).

    Uses the UUID4 specification.

    Returns
    -------
    str
        A UUID.
    """
    from uuid import uuid4

    return str(uuid4())


def check_valid_uuid(uuid_str: str) -> bool:
    """
    Check if the given uuid is valid.

    Parameters
    ----------
    uuid_str
        The uuid string to check.

    Returns
    -------
    bool
        True if the given uuid is valid, False otherwise.
    """
    with contextlib.suppress(ValueError):
        uuid_obj = uuid.UUID(uuid_str)
        if str(uuid_obj) == uuid_str:
            return True
    return False



"""A series of toy workflows that can be used for testing."""

from dataclasses import dataclass
from enum import Enum
from typing import Callable, NoReturn, Optional, Union

from jobflow import Job, JobConfig, Maker, OnMissing, Response, job


@job
def add(a, b):
    """Adds two numbers together and writes the answer to a file."""
    return a + b


@job
def always_fails() -> NoReturn:
    """A job that always fails."""
    raise RuntimeError("This job failed.")


@job
def write_file(n) -> None:
    with open("results.txt", "w") as f:
        f.write(str(n))


@job
def arithmetic(
    a: Union[float, list[float]],
    b: Union[float, list[float]],
    op: Optional[Callable] = None,
) -> Optional[float]:
    if op:
        return op(a, b)

    return None


@job
def check_env_var() -> str:
    import os

    return os.environ.get("TESTING_ENV_VAR", "unset")


@job(big_data="data")
def add_big(a: float, b: float):
    """Adds two numbers together and inflates the answer
    to a large list and tries to store that within
    the defined store.
    """
    result = a + b
    big_array = [result] * 5_000
    return Response({"data": big_array, "result": a + b})


@job(undefined_store="data")
def add_big_undefined_store(a: float, b: float):
    """Adds two numbers together and writes the answer to an artificially large file
    which is attempted to be stored in a undefined store.
    """
    result = a + b
    return Response({"data": [result] * 5_000, "result": result})


@job
def add_sleep(a, b):
    """Adds two numbers together and sleeps for "b" seconds."""
    import time

    time.sleep(b)
    return a + b


@job
def create_detour(detour_job: Job):
    """Create a detour based on the passed Job."""
    from jobflow import Flow

    return Response(detour=Flow(detour_job))


@job
def self_replace(n: int):
    """Create a replace Job with the same job n times."""
    if n > 0:
        return Response(replace=self_replace(n - 1))

    return n


@job
def ignore_input(a: int) -> int:
    """
    Can receive an input, but ignores it.

    Allows to test flows with failed parents
    """
    return 1


@job
def current_jobdoc():
    from jobflow_remote.jobs.run import CURRENT_JOBDOC

    return CURRENT_JOBDOC.job_doc


@job(config=JobConfig(resolve_references=False, on_missing_references=OnMissing.NONE))
def no_resolve(ref):
    """
    A job that does not resolve the reference in input
    """
    return ref


@job
def replace_and_stop_jobflow(a=1, b=1) -> Response[None]:
    from jobflow import Flow

    j1 = add(a, b)
    j2 = add(j1.output, 1)
    flow = Flow([j1, j2], output=j2.output)
    return Response(replace=flow, stop_jobflow=True)
    # return Response(replace=flow)


@job
def replace_and_stop_children(a: int = 1, b: int = 1) -> Response[None]:
    from jobflow import Flow

    j1 = add(a, b)
    j2 = add(j1.output, 1)
    flow = Flow([j1, j2], output=j2.output)
    return Response(replace=flow, stop_children=True)


@job
def stop_jobflow(x=None) -> Response[None]:
    return Response(stop_jobflow=True)


class TestEnum(Enum):
    A = "A"
    B = "B"


@dataclass
class EnumMaker(Maker):
    e: TestEnum = TestEnum.A
    name: str = "enum maker"

    @job
    def make(self):
        assert isinstance(self.e, TestEnum)
